package # hide from PAUSE
App::DBBrowser::Table::Substatements::Operators;

use warnings;
use strict;
use 5.014;

use List::MoreUtils qw( any uniq );

use Term::Choose         qw();
use Term::Form::ReadLine qw();

use App::DBBrowser::Auxil;
use App::DBBrowser::Table::Extensions;
#use App::DBBrowser::Table::ScalarFunctions::SQL;  # required


sub new {
    my ( $class, $info, $options, $d ) = @_;
    my $sf = {
        i => $info,
        o => $options,
        d => $d
    };
    bless $sf, $class;
}


sub build_having_col {
    my ( $sf, $sql, $clause, $aggr ) = @_;
    my $qt_aggr;
    if ( any { '@' . $_ eq $aggr } @{$sql->{aggr_cols}} ) {
        $qt_aggr = $aggr =~ s/^\@//r;
    }
    elsif ( $aggr eq 'COUNT(*)' ) {
        $qt_aggr = $aggr;
    }
    elsif ( any { $aggr eq $_ } @{$sf->{i}{avail_aggr}} ) {
        my $tc = Term::Choose->new( $sf->{i}{tc_default} );
        my $ax = App::DBBrowser::Auxil->new( $sf->{i}, $sf->{o}, $sf->{d} );
        my @pre = ( undef );
        if ( $sf->{o}{enable}{col_menu_addition} ) {
            push @pre, $sf->{i}{menu_addition};
        }
        $aggr =~ s/\(\S\)\z//;
        my $bu_having_stmt = $sql->{having_stmt};
        $sql->{having_stmt} .= ' ' . $aggr . "(";
        $qt_aggr          =       $aggr . "(";
        my $info = $ax->get_sql_info( $sql );
        $sql->{having_stmt} = $bu_having_stmt;
        my $qt_col;

        COLUMN: while( 1 ) {
            # Choose
            my $qt_col = $tc->choose(
                [ @pre, @{$sql->{cols}} ],
                { %{$sf->{i}{lyt_h}}, info => $info }
            );
            $ax->print_sql_info( $info );
            if ( ! defined $qt_col ) {
                return;
            }
            elsif ( $qt_col eq $sf->{i}{menu_addition} ) {
                my $ext = App::DBBrowser::Table::Extensions->new( $sf->{i}, $sf->{o}, $sf->{d} );
                my $complex_column = $ext->complex_unit( $sql, $clause );
                if ( ! defined $complex_column ) {
                    next COLUMN;
                }
                else {
                    $qt_col = $complex_column;
                }
            }
            $qt_aggr .= $qt_col . ")";
            last COLUMN;
        }
    }
    else { # SQ
        $qt_aggr = $aggr;
    }
    return $qt_aggr;
}


sub choose_and_add_operator {
    my ( $sf, $sql, $clause, $qt_col ) = @_;
    my $ax = App::DBBrowser::Auxil->new( $sf->{i}, $sf->{o}, $sf->{d} );
    my $tc = Term::Choose->new( $sf->{i}{tc_default} );
    my $stmt = $clause . '_stmt';
    my $args = $clause . '_args';
    my $menu_addition;
    my @operators_default;
    my @operators_complex;
    if ( $clause eq 'set' ) {
        $menu_addition = $sf->{i}{menu_addition};
        @operators_default = ( " = " );
        @operators_complex = ( " = " );
    }
    else {
        $menu_addition = '=' . $sf->{i}{menu_addition};
        @operators_default = grep { ! /\s(?:ANY|ALL)\z/ } @{$sf->{o}{G}{operators}};
        if ( $sf->{i}{driver} =~ /(?:Firebird|Informix)\z/ ) {
            @operators_default = uniq map { s/(?<=REGEXP)_i\z//; $_ } @operators_default;
        }
        @operators_complex = ( grep { ! /(?:\s%?col%?|REGEXP(?:_i)?|BETWEEN|NULL)\z/ } @{$sf->{o}{G}{operators}} );
    }
    if ( $sf->{o}{enable}{col_menu_addition} ) {
        unshift @operators_default, $menu_addition;
    }

    OPERATOR: while( 1 ) {
        my $is_complex_value;
        my $op;
        if ( @operators_default == 1 ) {
            $op = $operators_default[0];
        }
        else {
            my @pre = ( undef );
            my $info = $ax->get_sql_info( $sql );
            # Choose
            $op = $tc->choose(
                [ @pre, @operators_default ],
                { %{$sf->{i}{lyt_h}}, info => $info }
            );
            $ax->print_sql_info( $info );
            if ( ! defined $op ) {
                return;
            }
        }
        if ( $op eq $menu_addition ) {
            if ( @operators_complex == 1 ) {
                $op = $operators_complex[0];
            }
            else {
                my @pre = ( undef );
                my $info = $ax->get_sql_info( $sql );
                # Choose
                $op = $tc->choose(
                    [ @pre, @operators_complex ],
                    { %{$sf->{i}{lyt_h}}, info => $info, prompt => 'Operator:' }
                );
                $ax->print_sql_info( $info );
                if ( ! defined $op ) {
                    next OPERATOR;
                }
            }
            $is_complex_value = 1;
        }
        else {
            $is_complex_value = 0;
        }
        $op =~ s/^\s+|\s+\z//g;
        my $bu_stmt = $sql->{$stmt};
        my $ok = $sf->__add_operator( $sql, $clause, $qt_col, $op );
        if ( ! $ok ) {
            $sql->{$stmt} = $bu_stmt;
            next OPERATOR;
        }
        return $op, $is_complex_value;
    }
}


sub __add_operator {
    my ( $sf, $sql, $clause, $qt_col, $op ) = @_;
    my $ax = App::DBBrowser::Auxil->new( $sf->{i}, $sf->{o}, $sf->{d} );
    my $tc = Term::Choose->new( $sf->{i}{tc_default} );
    my $stmt = $clause . '_stmt';
    my $args = $clause . '_args';
    my $ok;
    $ax->print_sql_info( $ax->get_sql_info( $sql ) );
    if ( $op =~ /^(.+)\s(%?col%?)\z/ ) {
        $op = $1;
        my $arg = $2;
        $sql->{$stmt} .= ' ' . $op;
        my $qt_col;
        my @pre = ( undef );
        if ( $stmt eq 'having_stmt' ) {
            my @choices = ( @{$sf->{i}{avail_aggr}}, map( '@' . $_,  @{$sql->{aggr_cols}} ) );
            my $info = $ax->get_sql_info( $sql );
            # Choose
            my $aggr = $tc->choose(
                [ @pre, @choices ],
                { %{$sf->{i}{lyt_h}}, info => $info }
            );
            $ax->print_sql_info( $info );
            if ( ! defined $aggr ) {
                return;
            }
            $qt_col = $sf->build_having_col( $sql, $clause, $aggr );
        }
        else {
            my $info = $ax->get_sql_info( $sql );
            # Choose
            $qt_col = $tc->choose(
                [ @pre, @{$sql->{cols}} ],
                { %{$sf->{i}{lyt_h}}, info => $info, prompt => 'Col:' }
            );
            $ax->print_sql_info( $info );
        }
        if ( ! defined $qt_col ) {
            return;
        }
        if ( $arg !~ /%/ ) {
            $sql->{$stmt} .= ' ' . $qt_col;
        }
        else {
            if ( ! eval {
                require App::DBBrowser::Table::ScalarFunctions::SQL;
                my $fsql = App::DBBrowser::Table::ScalarFunctions::SQL->new( $sf->{i}, $sf->{o} );
                my @el = map { "'$_'" } grep { length $_ } $arg =~ /^(%?)(col)(%?)\z/g;
                my $qt_arg = $fsql->concatenate( \@el );
                $qt_arg =~ s/'col'/$qt_col/;
                $sql->{$stmt} .= ' ' . $qt_arg;
                1 }
            ) {
                $ax->print_error_message( $@ );
                return;
            }
        }
    }
    elsif ( $op =~ /REGEXP(_i)?\z/ ) {
        $sql->{$stmt} =~ s/ (?: (?<=\() | \s ) \Q$qt_col\E \z //x;
        my $do_not_match_regexp = $op =~ /^NOT/ ? 1 : 0;
        my $case_sensitive      = $op =~ /REGEXP_i\z/ ? 0 : 1;
        my $regex_op;
        if ( ! eval {
            $regex_op = $sf->_regexp( $qt_col, $do_not_match_regexp, $case_sensitive );
            1 }
        ) {
            $ax->print_error_message( $@ );
            return;
        }
        $regex_op =~ s/^\s// if $sql->{$stmt} =~ /\(\z/;
        $sql->{$stmt} .= $regex_op;
    }
    else {
        $sql->{$stmt} .= ' ' . $op;
    }
    $ax->print_sql_info( $ax->get_sql_info( $sql ) );
    return 1;
}


sub read_and_add_value {
    my ( $sf, $sql, $clause, $op, $is_complex_value ) = @_;
    my $ax = App::DBBrowser::Auxil->new( $sf->{i}, $sf->{o}, $sf->{d} );
    my $tr = Term::Form::ReadLine->new( $sf->{i}{tr_default} );
    my $stmt = $clause . '_stmt';
    my $args = $clause . '_args';
    if ( $is_complex_value ) {
        my $ext = App::DBBrowser::Table::Extensions->new( $sf->{i}, $sf->{o}, $sf->{d} );
        my $complex_value = $ext->complex_unit( $sql, $clause );
        if ( ! defined $complex_value ) {
            return;
        }
        if ( $op =~ /^(?:NOT\s)?IN\z/ && $complex_value !~ /^\s*\(.+\)\s*\z/ ) { # never ##
            $sql->{$stmt} .= '(' . $complex_value . ')';
        }
        else {
            $sql->{$stmt} .= ' ' . $complex_value;
        }
        return 1;
    }
    else {
        if ( $op =~ /^IS\s(?:NOT\s)?NULL\z/ ) {
            return 1;
        }
        elsif ( $op =~ /\s%?col%?\z/ ) {
            return 1;
        }
        elsif ( $op =~ /^(?:NOT\s)?IN\z/ ) {
            my $col_sep = '';
            $sql->{$stmt} .= ' (';
            my @bu;

            IN: while ( 1 ) {
                my $info = $ax->get_sql_info( $sql );
                # Readline
                my $value = $tr->readline(
                    'Value: ',
                    { info => $info }
                );
                $ax->print_sql_info( $info );
                if ( ! defined $value ) {
                    if ( @bu ) {
                        ( $sql->{$stmt}, $sql->{$args}, $col_sep ) = @{pop @bu};
                        next IN;
                    }
                    return;
                }
                if ( $value eq '' ) {
                    if ( $col_sep eq '' ) {
                        return;
                    }
                    $sql->{$stmt} .= ')';
                    return 1;
                }
                push @bu, [ $sql->{$stmt}, [ @{$sql->{$args}} ], $col_sep ];
                $sql->{$stmt} .= $col_sep . '?';
                push @{$sql->{$args}}, $value;
                $col_sep = ',';
            }
        }
        elsif ( $op =~ /^(?:NOT\s)?BETWEEN\z/ ) {
            my $info = $ax->get_sql_info( $sql );
            # Readline
            my $value_1 = $tr->readline(
                'Value 1: ',
                { info => $info }
            );
            $ax->print_sql_info( $info );
            if ( ! defined $value_1 ) {
                return;
            }
            $sql->{$stmt} .= ' ' . '?' . ' AND';
            push @{$sql->{$args}}, $value_1;
            $info = $ax->get_sql_info( $sql );
            # Readline
            my $value_2 = $tr->readline(
                'Value 2: ',
                { info => $info }
            );
            $ax->print_sql_info( $info );
            if ( ! defined $value_2 ) {
                return;
            }
            $sql->{$stmt} .= ' ' . '?';
            push @{$sql->{$args}}, $value_2;
            return 1;
        }
        elsif ( $op =~ /REGEXP(_i)?\z/ ) {
            push @{$sql->{$args}}, '...';
            my $info = $ax->get_sql_info( $sql );
            # Readline
            my $value = $tr->readline(
                'Pattern: ',
                { info => $info }
            );
            $ax->print_sql_info( $info );
            if ( ! defined $value ) {
                return;
            }
            $value = '^$' if ! length $value;
            pop @{$sql->{$args}};
            push @{$sql->{$args}}, $value;
            return 1;
        }
        else {
            my $value;
            my $info = $ax->get_sql_info( $sql );
            # Readline
            if ( $op =~ /^(?:NOT\s)?LIKE\z/ ) {
                $value = $tr->readline(
                    'Pattern: ',
                    { info => $info }
                );
            }
            else {
                $value = $tr->readline(
                    'Value: ',
                    { info => $info }
                );
            }
            $ax->print_sql_info( $info );
            if ( ! defined $value ) {
                return;
            }
            $sql->{$stmt} .= ' ' . '?';
            push @{$sql->{$args}}, $value;
            return 1;
        }
    }
}


sub _regexp {
    my ( $sf, $col, $do_not_match, $case_sensitive ) = @_;
    my $driver = $sf->{i}{driver};
    if ( $driver eq 'SQLite' ) {
        if ( $do_not_match ) {
            return sprintf " NOT REGEXP(?,%s,%d)", $col, $case_sensitive;
        }
        else {
            return sprintf " REGEXP(?,%s,%d)", $col, $case_sensitive;
        }
    }
    elsif ( $driver =~ /^(?:mysql|MariaDB)\z/ ) {
        if ( $do_not_match ) {
            return " $col NOT REGEXP ?"        if ! $case_sensitive;
            return " $col NOT REGEXP BINARY ?" if   $case_sensitive;
        }
        else {
            return " $col REGEXP ?"        if ! $case_sensitive;
            return " $col REGEXP BINARY ?" if   $case_sensitive;
        }
    }
    elsif ( $driver eq 'Pg' ) {
        if ( $do_not_match ) {
            return " ${col}::text !~* ?" if ! $case_sensitive;
            return " ${col}::text !~ ?"  if   $case_sensitive;
        }
        else {
            return " ${col}::text ~* ?" if ! $case_sensitive;
            return " ${col}::text ~ ?"  if   $case_sensitive;
        }
    }
    elsif ( $driver eq 'Firebird' ) {
        # SIMILAR TO
        # Unlike in some other languages, the pattern must match the entire
        # string in order to succeed — matching a substring is not enough.
        # wildcards: % and _
        if ( $do_not_match ) {
            return " $col NOT SIMILAR TO ? ESCAPE '#'";
        }
        else {
            return " $col SIMILAR TO ? ESCAPE '#'";
        }
    }
    elsif ( $driver =~ /^(?:DB2|Oracle)\z/ ) {
        if ( $do_not_match ) {
            return " NOT REGEXP_LIKE($col,?,'i')" if ! $case_sensitive;
            return " NOT REGEXP_LIKE($col,?,'c')" if   $case_sensitive;
        }
        else {
            return " REGEXP_LIKE($col,?,'i')" if ! $case_sensitive;
            return " REGEXP_LIKE($col,?,'c')" if   $case_sensitive;
        }
    }
    elsif ( $driver eq 'Informix' ) {
        if ( $do_not_match ) {
            return " $col NOT MATCHES ? ";
        }
        else {
            return " $col MATCHES ?";
        }
    }
}




1;


__END__
